import Vapor
import Fluent

struct AcronymsController: RouteCollection {
    func boot(routes: RoutesBuilder) throws {
        let acronymsRoute = routes.grouped("api", "acronyms")
        acronymsRoute.get(use: getAllHandler)
        acronymsRoute.post(use: createHandler)
        acronymsRoute.get(":acronymID", use: getHandler)
        acronymsRoute.delete(":acronymID", use: deleteHandler)
        acronymsRoute.put(":acronymID", use: updateHandler)
        acronymsRoute.get("search", use: searchHandler)
        acronymsRoute.get("first", use: firstHandler)
        acronymsRoute.get("sorted", use: sortedHandler)
        acronymsRoute.get(":acronymID", "user", use: getUserHandler)
        acronymsRoute.post(":acronymID", "categories", ":categoryID", use: attachCategoriesHandler)
        acronymsRoute.delete(":acronymID", "categories", ":categoryID", use: detachCategoriesHandler)
        acronymsRoute.get(":acronymID", "categories", use: getCategoriesHandler)
    }

    @Sendable
    func getAllHandler(_ req: Request) async throws -> [Acronym] {
        req.logger.info("Fetching all acronyms")
        return try await Acronym.query(on: req.db).all()
    }

    @Sendable
    func createHandler(_ req: Request) async throws -> Acronym {
        let data = try req.content.decode(AcronymCreateData.self)

        req.logger.info("Creating new acronym: \(data.short)")

        guard let user = try await User.find(data.userID, on: req.db) else {
            throw Abort(.notFound)
        }

        let acronym = Acronym(
            short: data.short, 
            long: data.long, 
            userID: user.id!
        )
        
        try await acronym.save(on: req.db)
        req.logger.info("Successfully created acronym: \(acronym.id?.uuidString ?? "")")

        return acronym
    }

    @Sendable
    func getHandler(_ req: Request) async throws -> Acronym {
        req.logger.info("Fetching acronym with ID: \(req.parameters.get("acronymID") ?? "")")
        guard let acronym = try await Acronym.find(req.parameters.get("acronymID"), on: req.db) else {
            throw TILError.acronymNotFound
        }
        return acronym
    }
    
    @Sendable
    func deleteHandler(_ req: Request) async throws -> HTTPStatus {
        req.logger.info("Deleting acronym with ID: \(req.parameters.get("acronymID") ?? "")")
        guard let acronym = try await Acronym.find(req.parameters.get("acronymID"), on: req.db) else {
            throw Abort(.notFound)
        }
        try await acronym.delete(on: req.db)
        req.logger.info("Successfully deleted acronym: \(acronym.id?.uuidString ?? "")")
        return .noContent
    }

    @Sendable
    func updateHandler(_ req: Request) async throws -> Acronym {
        let updatedData = try req.content.decode(AcronymCreateData.self)
        req.logger.info("Updating acronym with ID: \(req.parameters.get("acronymID") ?? "")")

        // 检查用户是否存在
        guard let _ = try await User.find(updatedData.userID, on: req.db) else {
            throw TILError.userNotFound
        }

        // 查找要更新的缩写
        guard let acronym = try await Acronym.find(
            req.parameters.get("acronymID"),
            on: req.db)
        else {
            throw TILError.acronymNotFound
        }

        // 检查是否存在重复
        let existingAcronym = try await Acronym.query(on: req.db)
            .filter(\.$short == updatedData.short)
            .filter(\.$id != acronym.id!)
            .first()

        if existingAcronym != nil {
            throw TILError.duplicateAcronym(updatedData.short)
        }

        acronym.short = updatedData.short
        acronym.long = updatedData.long
        try await acronym.save(on: req.db)

        req.logger.info("Successfully updated acronym: \(acronym.id?.uuidString ?? "")")
        return acronym
    }

    @Sendable
    func searchHandler(_ req: Request) async throws -> [Acronym] {
        guard let searchTerm = req.query[String.self, at: "term"] else {
            throw Abort(.badRequest)
        }
        req.logger.info("Searching acronyms with term: \(searchTerm)")
        return try await Acronym.query(on: req.db).group(.or) { or in
            or.filter(\.$short == searchTerm)
            or.filter(\.$long == searchTerm)
        }.all()
    }

    @Sendable
    func firstHandler(_ req: Request) async throws -> Acronym {
        req.logger.info("Fetching first acronym")
        guard let firstAcronym = try await Acronym.query(on: req.db).first() else {
            throw Abort(.notFound)
        }
        return firstAcronym
    }

    @Sendable
    func sortedHandler(_ req: Request) async throws -> [Acronym] {
        req.logger.info("Fetching sorted acronyms")
        return try await Acronym.query(on: req.db).sort(\.$short, .ascending).all()
    }

    @Sendable
    func getUserHandler(_ req: Request) async throws -> User {
        req.logger.info("Fetching user for acronym with ID: \(req.parameters.get("acronymID") ?? "")")
        guard let acronym = try await Acronym.find(req.parameters.get("acronymID"), on: req.db) else {
            throw Abort(.notFound)
        }
        return try await acronym.$user.get(on: req.db)
    }

    @Sendable
    func attachCategoriesHandler(_ req: Request) async throws -> HTTPStatus {
        req.logger.info("Attaching category to acronym with ID: \(req.parameters.get("acronymID") ?? "")")
        guard let acronym = try await Acronym.find(req.parameters.get("acronymID"), on: req.db) else {
            throw Abort(.notFound)
        }
        guard let category = try await Category.find(req.parameters.get("categoryID"), on: req.db) else {
            throw Abort(.notFound)
        }
        try await acronym.$categories.attach(category, on: req.db)
        req.logger.info("Successfully attached category: \(category.id?.uuidString ?? "") to acronym: \(acronym.id?.uuidString ?? "")")
        return .created
    }

    @Sendable
    func detachCategoriesHandler(_ req: Request) async throws -> HTTPStatus {
        req.logger.info("Detaching category from acronym with ID: \(req.parameters.get("acronymID") ?? "")")
        guard let acronym = try await Acronym.find(req.parameters.get("acronymID"), on: req.db) else {
            throw Abort(.notFound)
        }
        guard let category = try await Category.find(req.parameters.get("categoryID"), on: req.db) else {
            throw Abort(.notFound)
        }
        try await acronym.$categories.detach(category, on: req.db)
        req.logger.info("Successfully detached category: \(category.id?.uuidString ?? "") from acronym: \(acronym.id?.uuidString ?? "")")
        return .noContent
    }

    @Sendable
    func getCategoriesHandler(_ req: Request) async throws -> [Category] {
        req.logger.info("Fetching categories for acronym with ID: \(req.parameters.get("acronymID") ?? "")")
        guard let acronym = try await Acronym.find(req.parameters.get("acronymID"), on: req.db) else {
            throw Abort(.notFound)
        }
        return try await acronym.$categories.get(on: req.db)
    }
}

struct AcronymCreateData: Content {
    let short: String
    let long: String
    let userID: UUID
}